#!/bin/bash
#
#Name:mytask
#Ver:1.2
#Author:lykyl
#
#
#程序说明:
#简单计划任务
#
#实现功能：
#1、多任务并发执行，不会互相影响，采用锁机制避免单个任务的重叠执行。
#2、每个任务以单独脚本形式保存，相互独立。
#3、支持起始运行时间，如"2013/05/08"、"13:30"或“now”。并且支持给起始运行时间的修正值,比如"now+5m"表示当前时间的5分钟后执行（另外还实现了负数修正值，比-1h）。
#4、支持多种类型的运行周期设定，包括秒、分、时、天、周、月、年还有一次性任务。
#5、会根据任务执行间隔，自动设定休眠时间，主程序占用资源极小。
#6、任务可以是SHELL脚本,也可以是直接的linux命令。
#7、独立的任务配置文件,集中定义任务内容、起始时间和运行间隔。
#8、支持运行时重载任务配置文件，无需重启主程序就可以将调整后的任务加入计划。

BASEDIR=`dirname $0`    #程序所在目录
CONFFILE="$BASEDIR/mytask.conf"     #任务配置文件
LOGFILE="$BASEDIR/mytask.log"     #运行记录文件
PIDFILE="$BASEDIR/mytask.pid"     #pid存放文件
LOCKFILE="$BASEDIR/mytask.lock"     #锁文件
MISSTIMES=10    #任务过期运行阀值
DEFSLEEPTIME=60     #最短休眠时间片断
ReloadConf=0      #重载计划任务信息
ONEHOUR=3600
ONEDAY=86400
ONEWEEK=604800
INIT=-999999999
MAXTIME=99999999999

#防止程序重入创建锁文件
if [ -f $LOCKFILE ]; then
  exit 0
else
  touch $LOCKFILE
  #每次启动程序时覆写记录文件,如果需要保存历史运行记录,请自行修改程序
  echo "mytask start at "`date` >$LOGFILE
fi

#程序退出时清除所有锁文件
trap "rm -f $LOCKFILE;rm -f $BASEDIR/*_lock;echo 'exit';kill -15 $$" SIGINT EXIT
#利用自定义信号,平滑加载计划任务
trap "ReloadConf=1;echo 'reload'" SIGUSR1

declare -a aRunList
IFS=$'\n'

#获取任务配置信息
function GetConf()
{
  if ! [ -f $CONFFILE ]; then
    return
  fi
  TaskInfo=`grep -v '^\s*#' $CONFFILE|sed '/^\s*$/d'|uniq`
  aRunList=()
  for i in $TaskInfo
  do
    fn=`echo $i|awk -F, '{print $1}'`     #fn为任务内容
    startRun=`echo $i|awk -F, '{print $2}'`     #startRun为起始运行时间
    atime=`echo $i|awk -F, '{print $3}'`      #atime为运行周期
    if (( ${#fn} < 1 )) || (( ${#startRun} < 1 )) || (( ${#atime} < 1 )); then
      continue
    fi
    #利用MD5为每个任务生成唯一标识
    tn=`echo ${fn}|md5sum|awk '{print $1}'`
    
    #将初始执行时间全部统一为标准总秒数（+%s）
    startTime=${startRun%[+|-]*}
    startSec=`date -d "$startTime" +%s`
    fixTime=${startRun:${#startTime}:$[ ${#startRun} - ${#startTime} ]}
    case ${fixTime:$[ ${#fixTime} - 1]} in
      s|[0-9])
        #秒
        startSec=$[ $startSec + ${fixTime%s} ]
        ;;
      m)
        #分
        startSec=$[ $startSec + ${fixTime%m} * 60 ]
        ;;
      h)
        #时
        startSec=$[ $startSec + ${fixTime%h} * $ONEHOUR ]
        ;;
      d)
        #天
        startSec=$[ $startSec + ${fixTime%d} * $ONEDAY ]
        ;;
      w)
        #周
        startSec=$[ $startSec + ${fixTime%w} * $ONEWEEK ]
        ;;
      M)
        #月
        ty=`date -d $startTime +%y`
        tm=$[ `date -d $startTime +%m` + ${fixTime%M} ]
        td=$[ `date -d $startTime +%d` - 1 ]
        tt=`date -d $startTime +%T`
        if (( $tm > 12 )); then
          tm=$[ $tm % 12 ]
          ty=$[ $ty + $tm / 12 ]
        fi
        startSec=$[ `date -d "$ty-$tm-1 $tt" +%s` + $td * $ONEDAY ]
        ;;
      y)
        #年
        ty=$[ `date -d $startTime +%y` + ${fixTime%y} ]
        td=$[ `date -d $startTime +%j` - 1 ]
        tt=`date -d $startTime +%T`
        startSec=$[ `date -d "$ty-1-1 $tt" +%s` + $td * $ONEDAY ]
        ;;
    esac

    #计算任务执行间隔时间，将除单位为年和月以外的简隔时间统一为秒。
    #由于以年和月为单位的间隔时间要根据实际运行时间而定，所以不能预先计算。
    tp=s
    case ${atime:$[ ${#atime} - 1 ]} in
      s)
        addTime=${atime%s}
        ;;
      m)
        addTime=$[ ${atime%m} * 60 ]
        ;;
      h)
        addTime=$[ ${atime%h} * $ONEHOUR ]
        ;;
      d)
        addTime=$[ ${atime%d} * $ONEDAY ]
        ;;
      w)
        addTime=$[ ${atime%w} * $ONEWEEK ]
        ;;
      M)
        addTime=${atime%M}
        tp=M
        ;;
      y)
        addTime=${atime%y}
        tp=y
        ;;
      *)        
        addTime=$MAXTIME
        tp=0
        ;;
    esac

    #将初始化后的任务参数存入数组，供后续程序调用
    #任务参数以逗号分隔，分别为任务函数名、开始时间（标准总秒数）、运行间隔时间、间隔时间单位。
    #间隔时间单位为s、M、y，即秒、月、年。
    aRunList=(${aRunList[@]} "t$tn,$fn,$startSec,$addTime,$tp")
    echo "TempnamePre is t$tn,Task is $fn,start time is $startSec action type is $tp, action time is $addTime">>$LOGFILE;
  done
}

#程序主体
echo "$$">$PIDFILE
GetConf
while :
do
  #收到重新加载任务的信号后,调用GetConf函数并重置所有任务的下次执行时间
  if (( $ReloadConf >0  )); then
    ReloadConf=0
    echo "mytask restart at "`date` >>$LOGFILE;
    GetConf
    for i in ${aRunList[@]}
    do
      ntarg="${tn}_ntime"
      eval ${ntarg}=$startSec
    done
  fi
  if (( ${#aRunList[@]} <1 )); then
    exit
  fi
  IntervalTime=$INIT;
  nowTime=`date +%T`
  nowSec=`date +%s`
  for i in ${aRunList[@]}
  do
    tn=`echo $i|awk -F, '{print $1}'`
    fn=`echo $i|awk -F, '{print $2}'`
    startSec=`echo $i|awk -F, '{print $3}'`
    addTime=`echo $i|awk -F, '{print $4}'`
    tp=`echo $i|awk -F, '{print $5}'`
    if (( ${#fn} < 1 )) || (( ${#startSec} < 1 )) || (( ${#addTime} < 1 )) || (( ${#tp} < 1 )); then
      continue
    fi
    ntarg="${tn}_ntime"   #存放下次执行时间的变量名
    flagfile="${BASEDIR}/${tn}_lock"
    eval ${ntarg}=\${${ntarg}:=$startSec}
    eval tntarg=\$${ntarg}
    tdiff=$[ $nowSec - $tntarg ]
    echo "$fn diff:$tdiff" >>$LOGFILE
    #当前时间超过任务计划运行时间小于运行阀值时启动任务
    #为避免因某个任务执行时间过长超出此任务间隔时间而导致重入，
    #每个任务在执行时都会创建锁文件，并在任务执行完后删除。
    #为了保证多任务的并发性，每个任务都会以后台运行方式执行。
    if (( $tdiff >= 0 )); then
      if ! [ -e $flagfile ] && (( $tdiff < $MISSTIMES )) ; then
        { 
        touch $flagfile;
        echo "$fn start at "`date`\(`date +%s`\) >>$LOGFILE;
        eval $fn
        echo "$fn finished at "`date`\(`date +%s`\) >>$LOGFILE;
        rm -f $flagfile;
        } &
      else
        echo "$fn has skipped" >>$LOGFILE
      fi
      #根据间隔时间单位计算下一次任务执行的时间
      case $tp in
        s)
          addSec=$addTime
          ;;
        M)
          ty=`date +%y`
          tm=$[ `date +%m` + $addTime ]
          td=$[ `date +%d` - 1 ]
          if (( $tm > 12 )); then
            tm=$[ $tm % 12 ]
            ty=$[ $ty + $tm / 12 ]
          fi
          addSec=$[ `date -d "$ty-$tm-1 $nowTime" +%s` + $td * $ONEDAY ]
          ;;
        y)
          ty=$[ `date +%y` +$addTime ]
          td=$[ `date +%d` - 1 ]
          addSec=$[ `date -d "$ty-1-1 $nowTime" +%s` + $td * $ONEDAY ]
          ;;
        *)
          #将只执行一次的任务从任务数组中清除
          #由于将IFS设成了回车,所以移除数组项的代码写得很粗暴,下一步要考虑优化一下
          aL=()
          for j in ${aRunList[@]}
          do
            t=`echo $j|awk -F, '{print $1}'`
            if [ $t != $tn ]; then
              aL=(${aL[@]} $j)
            fi
          done
          aRunList=()
          for j in ${aL[@]}
          do
            aRunList=(${aRunList[@]} $j)
          done
          IntervalTime=0;
          break
          ;;
      esac
      tntarg=$[ $tntarg + ( $tdiff / $addSec ) * $addSec + $addSec ]
      eval ${ntarg}=$tntarg
      tdiff=$[ $nowSec - $tntarg ]
      echo "$fn next at $tntarg" >>$LOGFILE
    fi
    if (( $tdiff > $IntervalTime )) ; then      
      IntervalTime=$tdiff;
    fi
  done
  #主程序休眠到最近一次任务执行时
  if (( $IntervalTime <= $INIT )); then
    IntervalTime=1;
  fi
  if (( $IntervalTime < 0 )); then
    IntervalTime=$[ $IntervalTime * -1 ];
  fi
  echo "interval:${IntervalTime}s">>$LOGFILE
  #按DEFSLEEPTIME设定的值为最小休眠片断分段休眠，以防长时间休眠不能及时响应系统信号
  while (( $IntervalTime > 0 ))
  do
    if (( $IntervalTime < $DEFSLEEPTIME )); then
      stime=$IntervalTime
      IntervalTime=0
    else
      stime=$DEFSLEEPTIME
      IntervalTime=$[ $IntervalTime - $DEFSLEEPTIME ]
    fi
    sleep $stime
    if (( $ReloadConf >0 )); then
      ReloadConf=0
      echo "mytask restart at "`date` >>$LOGFILE;
      GetConf
      for i in ${aRunList[@]}
      do
        tn=`echo $i|awk -F, '{print $1}'`
        startSec=`echo $i|awk -F, '{print $3}'`
        ntarg="${tn}_ntime"
        eval ${ntarg}=$startSec
      done
    fi
  done
done
